/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * $Id: cdetach_SC.c,v 1.2 2001/01/03 09:55:55 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "config.h"
#include "global.h"
#include "cfs_SC_functions.h"

// cdetach_SC is a wrapper for the cfs function cdetach. 
// It an element of the cfs_SC program suite whose goal is to use 
// smartcards to store the cfs passphrase. 
//
// This program doesn't use smartcards but uses a file: the user cfs 
// mtab file which register the currently attached directories for a user.
// So, this wrapper uses and manages this file, and call the cdetach binary.
//
// This binary takes two options:
//    * -q: the program will not print any messages.
//    * -a: all the currently attached directories of the user will be
//          detached.
//
// If the -a option is not set, user must provide a clear name to detach.
////////////////////////////////////////////////////////////////////////////

// Internal functions
/////////////////////

// Print program usage informations
void usage(char *progName);

// Working part of the program
int cdetach_SC(char *clearDirName, int quiet);

int main(int argc, char **argv)
{
	int quiet = 0;				// Function is verbose by default
	int detachAll = 0;
	int rCode, result;
	char clearDirName[CFS_DIR_NAME_SIZE];

	// Get the arguments
	////////////////////

	while (1)
	{
		rCode = getopt(argc, argv, "aq");

		if (rCode == -1)
			break;

		switch (rCode)
		{
		case 'a':				// Detach all the currently attached directories
			detachAll = 1;
			break;

		case 'q':				// Be silent
			quiet = 1;
			break;

		case ':':				// Wrong option
			if (!quiet)
				usage(argv[0]);
			break;

		case '?':				// Wrong option
			if (!quiet)
				usage(argv[0]);
			break;
		}
	}

	if (!detachAll)
	{
		if (optind < argc)
		{
			// Copy the clear name argument
			strncpy(clearDirName, argv[optind], CFS_DIR_NAME_SIZE);
			clearDirName[strlen(argv[optind])] = '\0';
		}
		else if (!quiet)
			usage(argv[0]);		// Too less arguments
	}
	else if (optind < argc)
		if (!quiet)
			usage(argv[0]);		// Too much arguments

	if (detachAll)
	{
		// Detached all the directories stored 
		// in the user cfs mtab database.
		//////////////////////////////////////

		// Get an attached clear name
		rCode = give_dir_in_mtab(clearDirName, &result, quiet);

		if (rCode == EXIT_FAILURE)
		{
			if (!quiet)
				fprintf(stderr,
					"UNABLE to get an attached directory from cfs_SC_mtab\n");
			exit(EXIT_FAILURE);
		}

		if (!result)
			if (!quiet)
				printf("Nothing to detach\n");

		while (result)
		{
			if (!quiet)
				printf("Detaching %s\n", clearDirName);

			// Call the cdetach_SC function
			rCode = cdetach_SC(clearDirName, quiet);

			if (rCode == EXIT_FAILURE)
			{
				// We have to return and we don't try to detach the 
				// other directory...
				return (EXIT_FAILURE);
			}

			// Get another attached clear name
			rCode = give_dir_in_mtab(clearDirName, &result, quiet);

			if (rCode == EXIT_FAILURE)
			{
				if (!quiet)
					fprintf(stderr,
						"UNABLE to get an attached directory from cfs_SC_mtab\n");
				exit(EXIT_FAILURE);
			}
		}
	}

	else
	{
		// Just detach the single directory given in argument
		/////////////////////////////////////////////////////

		// Is this directory in user cfs mtab database?
		rCode = is_dir_in_mtab(clearDirName, &result, quiet);

		if (rCode == EXIT_FAILURE)
		{
			if (!quiet)
				fprintf(stderr, "UNABLE to use the user cfs mtab database\n");
			exit(EXIT_FAILURE);
		}

		if (!result)
		{
			if (!quiet)
				fprintf(stderr,
					"ERROR: the directory %s is not found in the user cfs mtab database\n",
					clearDirName);
			exit(EXIT_FAILURE);
		}

		// Call the cdetach_SC function 
		rCode = cdetach_SC(clearDirName, quiet);

		if (rCode == EXIT_FAILURE)
			exit(EXIT_FAILURE);
	}

	// Test if it's the last entry of cfs_SC_mtab
	rCode = is_mtab_empty(&result, quiet);

	if (rCode == EXIT_FAILURE)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to use the user cfs mtab database\n");
		exit(EXIT_FAILURE);
	}

	if (result)
	{
		// Ok, the user cfs mtab database is empty, we can delete it
		rCode = mtab_delete(quiet);

		if (rCode == EXIT_FAILURE)
		{
			if (!quiet)
				fprintf(stderr,
					"UNABLE to remove the user cfs mtab database\n");
			exit(EXIT_FAILURE);
		}
	}

	exit(EXIT_SUCCESS);
}

int cdetach_SC(char *clearDirName, int quiet)
{
	// This function do the detachement
	///////////////////////////////////

	int pid, rCode, status;

	// Call cdetach binary, we have to fork before call it
	pid = fork();

	if (pid == -1)
	{
		if (!quiet)
			fprintf(stderr, "UNABLE to fork: %s\n", strerror(errno));
		return (EXIT_FAILURE);
	}

	if (pid == 0)
	{
		// We are into the child process
		////////////////////////////////

		// If quiet is set, close stdout and stderr
		// Good method to silence the process!!
		if (quiet)
		{
			fclose(stdout);
			fclose(stderr);
		}

		// Exec cdetach binary
		rCode = execlp((CFS_BIN_DIR CDETACH), CDETACH, clearDirName, NULL);

		if (rCode == -1)
		{
			if (!quiet)
				fprintf(stderr, "UNABLE to execute %s: %s\n", CDETACH,
					strerror(errno));
			exit(EXIT_FAILURE);
		}
	}

	else
	{
		// We are still into the father process
		///////////////////////////////////////

		// Wait the end of the child
		rCode = waitpid(pid, &status, 0);

		if (WEXITSTATUS(status) != 0)
		{
			if (!quiet)
				fprintf(stderr, "UNABLE to detach the directory %s\n",
					clearDirName);
			return (EXIT_FAILURE);
		}

		// Update the user cfs mtab database in order to remove this deleted directory
		rCode = delete_dir_in_mtab(clearDirName, quiet);

		if (rCode == EXIT_FAILURE)
		{
			if (!quiet)
				fprintf(stderr,
					"UNABLE to delete the directory %s in the user cfs mtab database\n",
					clearDirName);
			return (EXIT_FAILURE);
		}
	}

	return (EXIT_SUCCESS);
}

void usage(char *progName)
{
	// Prints the program usage informations
	////////////////////////////////////////

	fprintf(stderr, "Usage: %s [-q] -a | name\n", progName);
	exit(EXIT_FAILURE);
}
