/*
 *
 * This program is written by
 * Vincent Guerin
 * and Ludovic Rousseau <ludovic.rousseau@gemplus.com>
 *
 * Code changed to fix a db open with db already open.
 * November 2001 - Domenico Taddei <domenico.taddei@proteam-group.it>
 *
 * Mon Nov  6 18:03:28 CET 2000
 * $Id: publickey_db_functions.c,v 1.5 2002/03/07 10:18:58 rousseau Exp $
 *
 */

/*
 * Copyright  2000 Gemplus
 * Av. du Pic de Bertagne - Parc d'Activits de Gmenos
 * BP 100 - 13881 Gmenos CEDEX
 * 
 * Everyone is allowed to redistribute and use this source  (source
 * code)  and binary (object code),  with or  without modification,
 * under some conditions:
 * 
 *  - Everyone  must  retain  and/or  reproduce the above copyright
 *    notice,  and the below  disclaimer of warranty and limitation
 *    of liability  for redistribution and use of these source code
 *    and object code.
 * 
 *  - Everyone  must  ask a  specific prior written permission from
 *    Gemplus to use the name of Gemplus.
 * 
 * DISCLAIMER OF WARRANTY
 * 
 * THIS CODE IS PROVIDED "AS IS",  WITHOUT ANY WARRANTY OF ANY KIND
 * (INCLUDING,  BUT  NOT  LIMITED  TO,  THE IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY  AND FITNESS FOR  A  PARTICULAR PURPOSE)  EITHER
 * EXPRESS OR IMPLIED.  GEMPLUS DOES NOT WARRANT THAT THE FUNCTIONS
 * CONTAINED  IN THIS SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR
 * THAT THE OPERATION OF IT WILL BE UNINTERRUPTED OR ERROR-FREE. NO
 * USE  OF  ANY  CODE  IS  AUTHORIZED  HEREUNDER EXCEPT UNDER  THIS
 * DISCLAIMER.
 * 
 * LIMITATION OF LIABILITY
 * 
 * GEMPLUS SHALL NOT BE LIABLE FOR INFRINGEMENTS OF  THIRD  PARTIES
 * RIGHTS. IN NO EVENTS, UNLESS REQUIRED BY APPLICABLE  LAW,  SHALL
 * GEMPLUS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES OF ANY CHARACTER  INCLUDING,
 * WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK STOPPAGE,
 * COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL OTHER DAMAGES OR
 * LOSSES, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. ALSO,
 * GEMPLUS IS  UNDER NO  OBLIGATION TO MAINTAIN,  CORRECT,  UPDATE,
 * CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS SOFTWARE.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <gdbm.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>

#include <rpcsvc/ypclnt.h>
#include <rpcsvc/yp_prot.h>

#include "config.h"
#include "global.h"
#include "publickey_db_functions.h"

// These functions are used to manage the public key database.
// This database stores the names of the registred users and
// their certified corresponding public key.
//////////////////////////////////////////////////////////////

// This structure is used to store the database elements
// in order to sort they.

typedef struct dbEntry
{
	char *user;
	char *key;
}
 *dbEntry;

// Internals functions
//////////////////////

// Comparison function for the RSA keys (used to sort the keys)
int compar_key_funct(dbEntry * E1, dbEntry * E2);

// Comparison function for user names (used to sort the names)
int compar_name_funct(char **name1, char **name2);

// Handler for some killing signals
void handler(int sig);

// Returns the name of the text file corresponding
// to the dbm database file.
int getTxtDbName(char *dbName, char *txtDbName);

// This function allow to copy the dbm database in a test file.
int dbm_to_txt(char *dbName);

int public_key_db_create_db(char *publicKeyDest, int new)
{
	// This function creates a public key database.
	// We use the dbm format database.
	///////////////////////////////////////////////

	GDBM_FILE dbf = NULL;

	// The dbm file must have the ".dbm" extension
	//////////////////////////////////////////////

	char *cp = NULL;

	// Seach the last "."
	cp = strrchr(publicKeyDest, '.');

	if (cp == NULL)
	{
		fprintf(stderr, "UNABLE to find the extension of the file %s\n",
			publicKeyDest);
		return (EXIT_FAILURE);
	}

	if (strcmp(cp, DBM_FILE_EXTENSION) != 0)
	{
		fprintf(stderr, "ERROR: extension of the file %s in not %s\n",
			publicKeyDest, DBM_FILE_EXTENSION);
		return (EXIT_FAILURE);
	}

	if (new)
	{
		// Creates a new database and so remove an existing database
		// having the same name.

		dbf = gdbm_open(publicKeyDest, DB_BLOC_SIZE, GDBM_NEWDB,
			OPEN_CREATE_DB_RIGHTS, NO_FATAL_FUNCT);
	}
	else
	{
		// If a database having the same name exists, an error is returned
		dbf = gdbm_open(publicKeyDest, DB_BLOC_SIZE, GDBM_WRCREAT,
			OPEN_CREATE_DB_RIGHTS, NO_FATAL_FUNCT);
	}

	if (dbf == NULL)
	{
		// Problem during the database creation
		fprintf(stderr, "UNABLE to create the public keys database(%s): %s\n",
			publicKeyDest, strerror(errno));
		return (EXIT_FAILURE);
	}

	gdbm_close(dbf);

	return (EXIT_SUCCESS);
}

int dbm_to_txt(char *dbName)
{
	// This function copies the dbm database conten in text file
	// The name of this file is the name of the dbm fiel without 
	// the ".dbm" extension.
	////////////////////////////////////////////////////////////

	int i, rCode;
	datum key;
	datum data;
	GDBM_FILE dbf;
	FILE *fd;
	char txtDbName[MAX_LEN_OF_DB_NAME];

	rCode = getTxtDbName(dbName, txtDbName);

	if (rCode == EXIT_FAILURE)
	{
		fprintf(stderr, "UNABLE to get the text database name\n");
		return (EXIT_FAILURE);
	}

	// Create the tmp text file
	fd = fopen(txtDbName, "w");
	chmod(txtDbName, 0444);

	if (fd == NULL)
	{
		fprintf(stderr, "UNABLE to open temporary file (%s): %s\n",
			txtDbName, strerror(errno));
		return (EXIT_FAILURE);
	}

	dbf = gdbm_open(dbName, DB_BLOC_SIZE, GDBM_READER,
		OPEN_CREATE_DB_RIGHTS, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		fprintf(stderr, "UNABLE to open public key database (%s): %s\n",
			dbName, strerror(errno));
		return (EXIT_FAILURE);
	}

	key = gdbm_firstkey(dbf);
	while (key.dptr)
	{
		char tmp[BUF_TMP_SIZE];

		data = gdbm_fetch(dbf, key);

		// Copy the username
		strncpy(tmp, key.dptr, key.dsize);
		tmp[key.dsize] = '\0';

		fprintf(fd, "%s", tmp);
		fprintf(fd, " ");

		// Copy the key
		for (i = 0; i < data.dsize; i++)
			fprintf(fd, "%02x", (unsigned char) data.dptr[i]);

		fprintf(fd, "\n");

		key = gdbm_nextkey(dbf, key);
	}

	// Close the files
	fclose(fd);
	gdbm_close(dbf);

	return (EXIT_SUCCESS);
}

int
publickey_db_add_user(char *username, unsigned char *publicKey,
	unsigned char *signature, char *publicKeyDest, int forcedFlag)
{
	int rCode;
	GDBM_FILE dbf = NULL;
	datum keyToAdd, dataToAdd;
	unsigned char data[RSA_KEY_SIZE_IN_BYTES + CERT_RSA_KEY_SIZE_IN_BYTES];

	// Create a new entry for this user
	// Return an error if the user is already regisred on the database
	//////////////////////////////////////////////////////////////////

	// Data are concatenation of the key and it's signature

	memcpy(data, publicKey, RSA_KEY_SIZE_IN_BYTES);
	memcpy(data + RSA_KEY_SIZE_IN_BYTES, signature,
		CERT_RSA_KEY_SIZE_IN_BYTES);

	keyToAdd.dptr = username;
	keyToAdd.dsize = strlen(username);

	dataToAdd.dptr = data;
	dataToAdd.dsize = RSA_KEY_SIZE_IN_BYTES + CERT_RSA_KEY_SIZE_IN_BYTES;

	dbf = gdbm_open(publicKeyDest, DB_BLOC_SIZE, GDBM_WRITER,
		OPEN_CREATE_DB_RIGHTS, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		fprintf(stderr, "UNABLE to open public key database (%s): %s\n",
			publicKeyDest, strerror(errno));
		return (EXIT_FAILURE);
	}

	if (forcedFlag)
		rCode = gdbm_store(dbf, keyToAdd, dataToAdd, GDBM_REPLACE);
	else
		rCode = gdbm_store(dbf, keyToAdd, dataToAdd, GDBM_INSERT);

	if (rCode != 0)				// User already registred on the database
		return (ERR_USR_ALREADY_IN_DB);

	gdbm_close(dbf);

	// Convert the dbm database to a text file
	rCode = dbm_to_txt(publicKeyDest);

	if (rCode == EXIT_FAILURE)
	{
		fprintf(stderr,
			"UNABLE to convert the dbm database to a text file\n");
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}

int publickey_db_user_in_db(char *username, char *publicKeyDest)
{
	int rCode;
	datum keyToSearch;
	GDBM_FILE dbf;

	// Search the user entry

	keyToSearch.dptr = username;
	keyToSearch.dsize = strlen(username);

	dbf = gdbm_open(publicKeyDest, DB_BLOC_SIZE, GDBM_READER,
		OPEN_CREATE_DB_RIGHTS, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		fprintf(stderr, "UNABLE to open public key database (%s): %s\n",
			publicKeyDest, strerror(errno));
		return (EXIT_FAILURE);
	}

	rCode = gdbm_exists(dbf, keyToSearch);

	gdbm_close(dbf);

	return (rCode);
}

int publickey_db_delete_user(char *username, char *publicKeyDest)
{
	// Delete the user entry
	////////////

	int rCode;
	datum keyToDelete;
	GDBM_FILE dbf;

	// Delete the user entry

	keyToDelete.dptr = username;
	keyToDelete.dsize = strlen(username);

	// Open the public keys database
	dbf = gdbm_open(publicKeyDest, DB_BLOC_SIZE, GDBM_WRITER,
		OPEN_CREATE_DB_RIGHTS, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		fprintf(stderr, "UNABLE to open public key database (%s): %s\n",
			publicKeyDest, strerror(errno));
		return (EXIT_FAILURE);
	}

	rCode = gdbm_delete(dbf, keyToDelete);

	if (rCode == -1)
		return (USR_NOT_IN_DB);

	gdbm_close(dbf);

	// Convert the dbm database to a text file
	rCode = dbm_to_txt(publicKeyDest);

	if (rCode == EXIT_FAILURE)
	{
		fprintf(stderr,
			"UNABLE to convert the dbm database to a text file\n");
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}

int publickey_db_list_user(char *publicKeyDest, char ***list, int *size)
{
	// List the users referenced in the database
	////////////////////////////////////////////

	int i = 0, nbUsers = 0;
	datum keyUser;
	GDBM_FILE dbf;

	// Open the public keys database
	dbf = gdbm_open(publicKeyDest, DB_BLOC_SIZE, GDBM_READER,
		OPEN_CREATE_DB_RIGHTS, NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		fprintf(stderr, "UNABLE to open public key database (%s): %s\n",
			publicKeyDest, strerror(errno));
		return (EXIT_FAILURE);
	}

	// First reading of the database to get the number of 
	// registred users in order to allocate the list memory.
	////////////////////////////////////////////////////////

	// Get the first key
	keyUser = gdbm_firstkey(dbf);

	while (keyUser.dptr)
	{
		nbUsers++;

		// Get the next key
		keyUser = gdbm_nextkey(dbf, keyUser);
	}

	// Set the list size
	*size = nbUsers;

	// Allocate the memory of the list
	*list = (char **) malloc(nbUsers * sizeof(char *));

	// Second reading to copy the user names in the list 
	// after have allocate memory for they.
	////////////////////////////////////////////////////

	// Get the first key
	keyUser = gdbm_firstkey(dbf);

	while (keyUser.dptr)
	{
		// Allocate the memory
		(*list)[i] = (char *) malloc((USER_NAME_MAX_SIZE + 1) * sizeof(char));

		// Copy the user name
		strncpy((*list)[i], keyUser.dptr, keyUser.dsize);
		(*list)[i][keyUser.dsize] = '\0';

		// Get the next key
		keyUser = gdbm_nextkey(dbf, keyUser);
		i++;
	}

	// Finally, sort the list by user using the qsort function 
	qsort(*list, nbUsers, sizeof(char *),
		(int (*)(const void *, const void *)) compar_name_funct);

	gdbm_close(dbf);

	return (EXIT_SUCCESS);
}

int publickey_db_verify_key(char *publicKeyDest, char *user_1, char *user_2)
{
	// Check the public key database in order to find duplicated keys
	/////////////////

	int keyDuplicated = 0, i = 0, nbUser = 0;
	datum keyUserIndex, dbKeyIndex;
	dbEntry *dbArray;
	GDBM_FILE dbf;

	// First, open the public keys database
	dbf =
		gdbm_open(publicKeyDest, DB_BLOC_SIZE, GDBM_READER, 0,
		NO_FATAL_FUNCT);

	if (dbf == NULL)
	{
		fprintf(stderr, "UNABLE to open public key database (%s): %s\n",
			publicKeyDest, strerror(errno));
		return (EXIT_FAILURE);
	}

	// First reading of the database to get the number of 
	// in order to allocate the memory of the array.
	/////////////////////////////////////////////////////

	keyUserIndex = gdbm_firstkey(dbf);

	while (keyUserIndex.dptr)
	{
		nbUser++;
		keyUserIndex = gdbm_nextkey(dbf, keyUserIndex);
	}

	// Allocate the array
	dbArray = (dbEntry *) malloc(nbUser * sizeof(dbEntry *));

	// Seconf reading of the databse to copy the database content 
	// in a dbEntry array.
	/////////////////////////////////////////////////////////////

	// Get the first key
	keyUserIndex = gdbm_firstkey(dbf);

	while (keyUserIndex.dptr)
	{
		dbKeyIndex = gdbm_fetch(dbf, keyUserIndex);

		// Allocate the memory for the structure and its fields
		dbArray[i] = (dbEntry) malloc(sizeof(struct dbEntry));
		dbArray[i]->user =
			(char *) malloc((keyUserIndex.dsize + 1) * sizeof(char));
		dbArray[i]->key =
			(char *) malloc(RSA_KEY_SIZE_IN_BYTES * sizeof(char));

		// Copy the data
		strncpy(dbArray[i]->user, keyUserIndex.dptr, keyUserIndex.dsize);
		(dbArray[i]->user)[keyUserIndex.dsize] = '\0';
		strncpy(dbArray[i]->key, dbKeyIndex.dptr, RSA_KEY_SIZE_IN_BYTES);

		// Get the next key
		keyUserIndex = gdbm_nextkey(dbf, keyUserIndex);
		i++;
	}

	gdbm_close(dbf);

	// Now we have to sort the array by public keys.
	// We just use the qsort function.
	qsort(dbArray, nbUser, sizeof(dbEntry *),
		(int (*)(const void *, const void *)) compar_key_funct);

	// Now, we use the public keys sorted array to find out the first doubles
	/////////////////////////////////////////////////////////////////////////

	for (i = 0; i < nbUser - 1; i++)
		if (strncmp(dbArray[i]->key, dbArray[i + 1]->key,
				RSA_KEY_SIZE_IN_BYTES) == 0)
		{
			// Double found, store the corresponding registred users 
			// having this same key.
			////////////////////////////////////////////////////////

			sprintf(user_1, "%s", dbArray[i]->user);
			sprintf(user_2, "%s", dbArray[i + 1]->user);

			keyDuplicated++;
			goto OUT;
		}

  OUT:
	// Free the array

	for (i = 0; i < nbUser; i++)
	{
		free((dbArray[i])->user);
		free((dbArray[i])->key);
	}

	free(dbArray);

	return keyDuplicated;
}

int compar_key_funct(dbEntry * E1, dbEntry * E2)
{
	// Comparison function used by qsort()
	// Just uses the string comparison function on the keys.
	return strncmp((*E1)->key, (*E2)->key, RSA_KEY_SIZE_IN_BYTES);
}

int compar_name_funct(char **name1, char **name2)
{
	// Another comparison function used by qsort()
	// Just uses the string comparison function.
	return strncmp(*name1, *name2, USER_NAME_MAX_SIZE);
}

int test_db(char *dbFileName)
{
	// This function returns EXIT_SUCCESS if the specified 
	// database exists, EXIT_FAILURE else.
	//////////////////////////////////////////////////////

	int rCode;
	struct stat finfo;

	// Test the existence of th specified database file
	rCode = lstat(dbFileName, &finfo);

	if ((rCode == -1) && errno != ENOENT)
	{
		// Unable to stat the file
		fprintf(stderr, "UNABLE to stat the %s file: %s\n",
			dbFileName, strerror(errno));
		return (EXIT_FAILURE);
	}

	else if ((rCode == -1) && errno == ENOENT)
	{
		// The file doesn't exist
		return (EXIT_FAILURE);
	}

	else
		return (EXIT_SUCCESS);	// Database exists
}

int test_lock()
{
	// This function returns EXIT_SUCCESS and installs the lock file
	// is we are the only process to modify the public key database.
	// Returns EXIT_FAILURE if not.
	////////////////////////////////////////////////////////////////

	int fd;

	// Install the killing signals handler in order to quit properly
	signal(SIGINT, handler);
	signal(SIGQUIT, handler);

	// Verify if there is no smart session tool process running and 
	// create the lock file using O_EXCL option. 
	///////////////////////////////////////////////////////////////

	fd = open(SST_LOCK_FILE, O_WRONLY | O_CREAT | O_EXCL,
		SST_LOCK_FILE_RIGHTS);

	if (fd == -1)
	{
		if (errno == EEXIST)
		{
			// File already exists, an other identical process is running
			fprintf(stderr,
				"Another smart session tools process is running!!\n");
			fprintf(stderr,
				"If it's abnormal remove manually the " SST_LOCK_FILE
				" file\n");
		}
		else
		{
			// Other problem
			fprintf(stderr,
				"Unable to create the lock file" SST_LOCK_FILE ": %s\n",
				strerror(errno));
			fprintf(stderr, "Are you root?\n");
		}

		return (EXIT_FAILURE);
	}

	// Ok we are alone
	return (EXIT_SUCCESS);
}

void handler(int sig)
{
	// If process exits abnormaly, we have to remove the lock file
	unlink(SST_LOCK_FILE);

	exit(EXIT_SUCCESS);
}

int getTxtDbName(char *dbName, char *txtDbName)
{
	// This function returns the name of the text file corresponding
	// to the dbm database file.
	////////////////////////////////////////////////////////////////

	// The name is the dbm name without the ".dbm" extension.

	char *cp = NULL;

	strncpy(txtDbName, dbName, MAX_LEN_OF_DB_NAME);

	// Seach the last "."
	cp = strrchr(txtDbName, '.');

	if (cp == NULL)
	{
		fprintf(stderr, "UNABLE to find the extension of the file %s\n",
			dbName);
		return (EXIT_FAILURE);
	}

	if (strcmp(cp, DBM_FILE_EXTENSION) != 0)
	{
		fprintf(stderr, "ERROR: extension of the file %s in not %s\n",
			dbName, DBM_FILE_EXTENSION);
		return (EXIT_FAILURE);
	}

	// Truncate foobar.DBM_FILE_EXTENSION into foobar
	*cp = '\0';

	return (EXIT_SUCCESS);
}
